theory Safety_Proof_Parser
imports
  Safety_Checker
  "$AFP/XML/Xmlt"
begin

datatype ('v,'l,'t) safety_cert_problem = Safety_Cert_Problem "('v,'l,'t)lin_prog_impl \<times> 'l" "('v,'l,'t)art_impl"

context fixes
  node :: "'l xmlt" and
  trans_id :: "'t xmlt"
begin

context
  fixes variable :: "'v xmlt"
begin

definition "constant" :: "int xmlt" where
  "constant = Xmlt.int ''constant''"

definition lin_exp :: "'v lin_exp xmlt" where
  "lin_exp = Xmlt.many1 ''term'' constant (Xmlt.pair ''addend'' constant variable (\<lambda> c x. (x,c))) Lin_Exp"

definition lin_atom :: "'v lin_atom xmlt" where
  "lin_atom = Xmlt.singleton ''atom'' lin_exp Exp_Le_0"

definition formula :: "'v formula xmlt" where
  "formula = Xmlt.singleton ''formula'' (Xmlt.many ''conjunction'' lin_atom id) Conjunction"
end

context
  fixes tvariable :: "'v trans_var xmlt"
  fixes variable :: "'v xmlt"
begin

abbreviation "transition_formula \<equiv> formula tvariable"

definition transition :: "('t \<times> ('v,'l)transition) xmlt" where
  "transition = Xmlt.tuple4 ''transition'' 
    trans_id 
    (Xmlt.singleton ''source'' node id)
    (Xmlt.singleton ''target'' node id)
    transition_formula
    (\<lambda> i s t f. (i, Transition s f t))
    "

definition lin_prog :: "('v,'l,'t)lin_prog_impl xmlt" where
  "lin_prog = Xmlt.many1 ''program'' 
     (Xmlt.singleton ''initial'' node id)
     transition
     Lin_Prog_Impl"

definition safety_problem :: "(('v,'l,'t)lin_prog_impl \<times> 'l)xmlt" where
  "safety_problem = Xmlt.pair ''safetyProgram''
     lin_prog
     (Xmlt.singleton ''error'' node id)
     Pair"

definition art_node_id :: "art_node xmlt" where
  "art_node_id = Xmlt.text ''artNodeId''"

definition entailment_info :: "entailment_info xmlt" where
  "entailment_info = Xmlt.many ''hints''
    (Xmlt.singleton ''linearImplicationHint''
       (Xmlt.many ''linearCombination'' constant (map nat)) id) Entailment_Hint"

definition art_node :: "('v,'l,'t)art_node_impl xmlt" where
  "art_node = Xmlt.tuple4 ''artNode''
     art_node_id
     (formula variable)
     node
     (Xmlt.options [
       (''children'', Xmlt.many ''children'' 
         (Xmlt.triple ''child'' trans_id art_node_id entailment_info (\<lambda> t n h. (t,n,h))) Children_Edge),
       (''coverEdge'', Xmlt.pair ''coverEdge'' art_node_id entailment_info Cover_Edge)])
     Art_Node"

definition art_safety_proof :: "('v,'l,'t)art_impl xmlt" where
  "art_safety_proof = Xmlt.triple ''impact''
     art_node_id
     (Xmlt.many ''artNodes'' art_node id)
     (Xmlt.many ''errorHints'' (Xmlt.pair ''errorHint'' art_node_id entailment_info Pair) id)
     (\<lambda> init ns hs. \<lparr>art_impl.initial_node = init, art_nodes = ns, error_hints = hs\<rparr>)"

definition safety_proof :: "('v,'l,'t)art_impl xmlt" where
  "safety_proof = Xmlt.singleton ''safetyProof'' art_safety_proof id"

definition safety_certification_problem :: "('v,'l,'t)safety_cert_problem xmlt" where
  "safety_certification_problem = Xmlt.pair ''safetyProblem'' safety_problem safety_proof Safety_Cert_Problem"
end
end

type_synonym node = string
type_synonym trans_id = string
type_synonym variable = string

definition node :: "node xmlt" where
  "node = Xmlt.singleton ''node''
    (Xmlt.text ''nodeIdentifier'') id"

definition trans_id :: "trans_id xmlt" where
  "trans_id = Xmlt.text ''transitionId''"

definition variable_name :: "variable xmlt" where
  "variable_name = Xmlt.singleton ''variableName''
     (Xmlt.text ''variableIdentifier'') id"

definition variable :: "variable xmlt" where
  "variable = Xmlt.singleton ''variable'' variable_name id"

definition trans_variable :: "variable trans_var xmlt" where
  "trans_variable = Xmlt.singleton ''variable''
    (Xmlt.options [
      (''variableName'', Xmlt.change variable_name Pre),
      (''post'', Xmlt.singleton ''post'' variable_name Post)
    ]) id"

fun drop_xml_attributes :: "xml \<Rightarrow> xml" where
  "drop_xml_attributes (XML name atts cs) = (XML name [] cs)"
| "drop_xml_attributes x = x"

primrec
  xmldoc_safety_cert_problem ::
    "xmldoc \<Rightarrow> string +\<^sub>\<bottom> (variable, node, trans_id) safety_cert_problem"
where
  "xmldoc_safety_cert_problem (XMLDOC header xml) = 
    (safety_certification_problem node trans_id trans_variable variable 
      (drop_xml_attributes xml))"


definition
  parse_safety_cert_problem ::
    "string \<Rightarrow> string +\<^sub>\<bottom> (variable, node, trans_id) safety_cert_problem"
where
  "parse_safety_cert_problem s =
    (case doc_of_string s of
      Inl x \<Rightarrow> error x
    | Inr y \<Rightarrow> xmldoc_safety_cert_problem y)"

end
